import numpy as np
from pymoo.factory import get_performance_indicator
import scipy.stats
import pandas as pd

#This is an experimental file that might be used for comparing algorithms
#Pandas was not included in the requirements

def get_igd_history(problem, result):
    
    igd_function = get_performance_indicator("igd", problem.pareto_front(), norm_by_dist=True)
    
    igd_list = []
    
    for gen in result.history:
        
        F = gen.opt.get("F")
        igd = igd_function.do(F)
        igd_list.append(igd)
    
    return igd_list

def get_hv_history(problem, result):
    
    refpoint = problem.pareto_front().max(axis=0)
    nadir = refpoint
    ideal = problem.pareto_front().min(axis=0)
    hv_function = get_performance_indicator("hv", refpoint, zero_to_one=True,
                                            nadir=nadir, ideal=ideal)
    
    hv_list = []
    
    for gen in result.history:
        
        F = gen.opt.get("F")
        hv = hv_function.do(F)
        hv_list.append(hv)
    
    return hv_list

def get_experiment_metric(results, problem, metric):
    
    columns = np.arange(0, len(results), 1) + 1
    
    metrics = {}
    
    if metric == "igd":
        perf = get_igd_history
    elif metric == "hv":
        perf = get_hv_history
    else:
        raise KeyError("Unknown metric")
    
    for j, res in enumerate(results):
        
        metrics[columns[j]] = perf(problem, res)
    
    return metrics

class StatsSignificance:
    
    def __init__(self, lesser_is_better=True):
        
        self.lesser_is_better = lesser_is_better 
        self.best_median = None
        self.best_values = None
        self.best_label = None
        self.results = pd.DataFrame(columns=["median", "mean", "std",
                                             "shapiro", "levene", "t-test", "wilcoxon"])
    
    def get_best_from_dict(self, candidates):
        
        for key, value in candidates.items():
            
            if self.best_values is None:
                self.best_label = key
                self.best_values = value.iloc[-1, :]
                self.best_median = value.iloc[-1, :].median()
            else:
                if self.lesser_is_better and (value.iloc[-1, :].median() < self.best_median):
                    self.best_label = key
                    self.best_values = value.iloc[-1, :]
                    self.best_median = value.iloc[-1, :].median()
                elif (not self.lesser_is_better) and (value.iloc[-1, :].median() > self.best_median):
                    self.best_label = key
                    self.best_values = value.iloc[-1, :]
                    self.best_median = value.iloc[-1, :].median()
                else:
                    pass
    
    def compare(self, candidate, label=None):
        
        _, pval = scipy.stats.shapiro(candidate)
        shapiro = "* (%.3f)" % pval if pval >= 0.01 else ""

        _, pval = scipy.stats.levene(self.best_values, candidate)
        levene = "* (%.3f)" % pval if pval >= 0.05 else ""
        
        if self.lesser_is_better:
            alternative = "greater"
        else:
            alternative = "smaller"

        _, pval = scipy.stats.ttest_ind(candidate, self.best_values)
        ttest = "* (%.3f)" % pval if pval >= 0.05 else ""

        if len(self.best_values) == len(candidate):
            _, pval = scipy.stats.wilcoxon(candidate, self.best_values, zero_method="zsplit")
            wilcoxon = "* (%.3f)" % pval if pval >= 0.05 else ""
        else:
            wilcoxon = "x"
        
        if label is None:
            label = self.results.shape[0] + 1
        
        median_, mean_, std_ = candidate.median(), candidate.mean(), candidate.std()
        
        self.results.loc[label, :] = [median_, mean_, std_, shapiro, levene, ttest, wilcoxon]



import numpy as np
from pymoo.algorithms.moo.nsga2 import calc_crowding_distance as _calc_crowding_distance
from pymoo.util.randomized_argsort import randomized_argsort
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.core.survival import Survival, split_by_feasibility
from pymoo.core.population import Population
from pymoo.util.misc import find_duplicates
from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import NearestNeighbors

#Redefine function to be able to support more kwargs
def calc_crowding_distance(F, filter_out_duplicates=True, **kwargs):
    return _calc_crowding_distance(F, filter_out_duplicates=True)


def get_crowding_function(label):

    if label == "cd":
        fun = FunctionalDiversity(calc_crowding_distance)
    elif label == "ce":
        fun = FunctionalDiversity(calc_crowding_entropy)
    #Inefficient form based on full pairwise distances matrix
    elif label == "mnn_bak":
        fun = FunctionalDiversity(_calc_mnn)
    elif label == "mnn":
        fun = MNNDiversity(fast_mode=False)
    elif label == "mnn-fast":
        fun = MNNDiversity(fast_mode=True)
    elif label == "2nn":
        fun = MNNDiversity(fast_mode=False, twonn=True)
    elif label == "2nn-fast":
        fun = MNNDiversity(fast_mode=True, twonn=True)
    else:
        raise KeyError("Crwoding function not defined")
    return fun

class RankSurvival(Survival):

    def __init__(self,
                 nds=None,
                 rule="full",
                 crowding_func="cd"):
        
        """
        A generalization of the NSGA-II survival operator that ranks individuals by dominance criteria
        and sorts the last front by some crowding metric.

        Parameters
        ----------
        
        nds : str or None, optional
            Pymoo type of non-dominated sorting. Defaults to None.

        rule : str, optional
            Rule to remove individuals exceeding popsize. Options are:
            
                - 'full'
                - 'sqrt'
                - 'single'
            
            The rule 'full' corresponds to the original version, in which all individuals exceeding the front are removed at onde. Other options remove individuals recursively and recalculate metrics at each iteration, which helps improving diversity, although is more computationally expensive. Defaults to 'full'.
            
        crowding_func : str or callable, optional
            Crowding metric. Options are:
            
                - 'cd': crowding distances
                - 'ce': crowding entropy
                - 'mnn': M-Neaest Neighbors
                - '2nn': 2-Neaest Neighbors
                - 'mnn-fast': M-Neaest Neighbors Bulk Removal
                - '2nn-fast': 2-Neaest Neighbors Bulk Removal
                
            If callable, it has the form ``fun(F, filter_out_duplicates=None, n_remove=None, **kwargs)``
            in which F (n, m) and must return metrics in a (n,) array. The options 'cd' and 'ce' are recommended for two-objective problems, whereas 'mnn' and '2nn' for many objective.
            When using either 'mnn' or '2nn', individuals are already eliminated in a 'single' manner, 
            therefore 'rule' is ignored. To bulk removal, chose '-fast' variants.
            Defaults to 'cd'.
        """
        if crowding_func in ("mnn", "2nn"):
            rule = "full"
        
        if not hasattr(crowding_func, "__call__"):
            crowding_func = get_crowding_function(crowding_func)

        super().__init__(filter_infeasible=True)
        self.nds = nds if nds is not None else NonDominatedSorting()
        self.rule = rule
        self.crowding_func = crowding_func
        
    def _do(self,
            problem,
            pop,
            *args,
            n_survive=None,
            **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(float, copy=False)

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = self.nds.do(F, n_stop_if_ranked=n_survive)

        for k, front in enumerate(fronts):

            # current front sorted by crowding distance if splitting
            while len(survivors) + len(front) > n_survive:
                
                #Define how many will be removed
                if self.rule == "full":
                    n_remove = len(survivors) + len(front) - n_survive
                elif self.rule == "single":
                    n_remove = 1
                elif self.rule == "sqrt":
                    n_remove = int(np.sqrt(len(survivors) + len(front) - n_survive))
                
                # re-calculate the crowding distance of the front
                crowding_of_front = self.crowding_func.do(
                    F[front, :], n_remove=n_remove
                    )
                I = randomized_argsort(crowding_of_front, order='descending', method='numpy')
                
                I = I[:-n_remove]
                front = front[I]

            # otherwise take the whole front unsorted
            else:
                # calculate the crowding distance of the front
                crowding_of_front = self.crowding_func.do(
                    F[front, :], n_remove=0
                    )

            # save rank and crowding in the individual class
            for j, i in enumerate(front):
                pop[i].set("rank", k)
                pop[i].set("crowding", crowding_of_front[j])

            # extend the survivors by all or selected individuals
            survivors.extend(front)

        return pop[survivors]


class ConstrainedRankSurvival(Survival):
    
    def __init__(self, nds=None, ranking=None):
        """
        The Rank and Crowding survival approach for handling constraints proposed on
        GDE3 by Kukkonen, S. & Lampinen, J. (2005).

        Parameters
        ----------
        
        nds : str or None, optional
            Pymoo type of non-dominated sorting. Defaults to None.
            
        ranking : Survival, optional
            Basic survival operator that splts by feasibility. 
            Feasible and infeasible solutions are ranked by nds separately. Defaults to None,
            which creates a RankSurvival instance.
        """
        super().__init__(filter_infeasible=False)
        self.nds = nds if nds is not None else NonDominatedSorting()
        self.ranking = ranking if ranking is not None else RankSurvival()
    
    def _do(self,
            problem,
            pop,
            *args,
            n_survive=None,
            **kwargs):

        if n_survive is None:
            n_survive = len(pop)

        n_survive = min(n_survive, len(pop))

        #If the split should be done beforehand
        if problem.n_constr > 0:

            #Split by feasibility
            feas, infeas = split_by_feasibility(pop, eps=0.0, sort_infeasbible_by_cv=True)

            #Obtain len of feasible
            n_feas = len(feas)

            #Assure there is at least_one survivor
            if n_feas == 0:
                survivors = Population()
            else:
                survivors = self.ranking.do(problem, pop[feas], *args, n_survive=min(len(feas), n_survive), **kwargs)

            #Calculate how many individuals are still remaining to be filled up with infeasible ones
            n_remaining = n_survive - len(survivors)

            #If infeasible solutions need to be added
            if n_remaining > 0:
                
                #Constraints to new ranking
                G = pop[infeas].get("G")
                G = np.maximum(G, 0)
                
                #Fronts in infeasible population
                infeas_fronts = self.nds.do(G, n_stop_if_ranked=n_remaining)
                
                #Iterate over fronts
                for k, front in enumerate(infeas_fronts):

                    #Save ranks
                    pop[infeas][front].set("cv_rank", k)

                    #Current front sorted by CV
                    if len(survivors) + len(front) > n_survive:
                        
                        #Obtain CV of front
                        CV = pop[infeas][front].get("CV").flatten()
                        I = randomized_argsort(CV, order='ascending', method='numpy')
                        I = I[:(n_survive - len(survivors))]

                    #Otherwise take the whole front unsorted
                    else:
                        I = np.arange(len(front))

                    # extend the survivors by all or selected individuals
                    survivors = Population.merge(survivors, pop[infeas][front[I]])

        else:
            survivors = self.ranking.do(problem, pop, *args, n_survive=n_survive, **kwargs)

        return survivors
    

class CrowdingDiversity:
    
    def do(self, F, filter_out_duplicates=True, n_remove=None):
        return self._do(F, filter_out_duplicates=filter_out_duplicates, n_remove=n_remove)
    
    def _do(self, F, filter_out_duplicates=True, n_remove=None):
        pass


class FunctionalDiversity(CrowdingDiversity):
    
    def __init__(self, function=None) -> None:
        self.function = function
        super().__init__()
    
    def _do(self, F, filter_out_duplicates=True, **kwargs):
        return self.function(F, filter_out_duplicates=filter_out_duplicates, **kwargs)
    

class MNNDiversity(CrowdingDiversity):
    
    def __init__(self, fast_mode=False, twonn=False, **kwargs) -> None:
        """
        Kukkonen, S. & Deb, K., 2006. A fast and effective method for pruning of non-dominated solutions in many-objective problems. In: Parallel problem solving from nature-PPSN IX. Berlin: Springer, pp. 553-562

        Parameters
        ----------
        
        fast_mode : bool, optional
            Eliminate all individuals at once. Defaults to False.
            
        twonn : bool, optional
            Use 2-NN instead of than M-NN. Defaults to False.
        """
        super().__init__()
        self.fast_mode = fast_mode
        self.twonn = twonn
    
    def _do(self, F, filter_out_duplicates=True, n_remove=None, **kwargs):
        
        n_points, n_obj = F.shape

        if n_points <= n_obj:
            return np.full(n_points, np.inf)

        else:

            if filter_out_duplicates:
                # filter out solutions which are duplicates - duplicates get a zero finally
                is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-32)))[0]
            else:
                # set every point to be unique without checking it
                is_unique = np.arange(n_points)

            # index the unique points of the array
            _F = F[is_unique].copy()
            
            #Break if many duplicates
            if _F.shape[0] <= n_obj:
                return np.full(n_points, np.inf)

            # calculate the norm for each objective - set to NaN if all values are equal
            norm = np.max(_F, axis=0) - np.min(_F, axis=0)
            norm[norm == 0] = 1.0
            
            #F normalized
            _F = (_F - _F.min(axis=0)) / norm
            
            #Create the heap H
            H = np.arange(_F.shape[0])
            
            #Find best at each function and assign np.inf
            ideal_idx = np.argmin(_F, axis=0)
            
            #Create knn instance
            if self.twonn:
                knn = NearestNeighbors(n_neighbors=3)
            else:
                knn = NearestNeighbors(n_neighbors=n_obj + 1)
            
            #Fit knn to front
            knn.fit(_F)
        
            #Get distances and indexes
            _D, _mnn = knn.kneighbors(X=_F, return_distance=True)
            D, Mnn = _D[:, 1:], _mnn[:, 1:]
            
            #Get d metrics
            d = np.prod(D, axis=1)
            d[ideal_idx] = np.inf
            
            #Avoid loop if in fast mode
            if self.fast_mode:
                
                #Initialize as zeros
                d2 = np.zeros(n_points)
                d2[is_unique] = d

                return d2
            
            #Elements to keep
            n_keep = _F.shape[0] - n_remove
            
            _iter = 0
            
            #Remove elements from heap recursively
            while ((H.shape[0] > n_keep) and (H.shape[0] > n_obj + 1)):
                
                #Most crowded element in H
                w = H[np.argmin(d[H])]
                
                #Remove w from H
                H = H[H != w]
                
                #Find neighbors of w
                _wnn = np.any(Mnn[H] == w, axis=1)
                wnn = H[_wnn]
                
                #If any items are to be re-calculated
                if np.any(_wnn):
                    
                    #Fit knn to remaining individuals
                    knn.fit(_F[H])
                
                    #Get distances and indexes
                    _D, _mnn = knn.kneighbors(X=_F[wnn], return_distance=True)
                    D[wnn], Mnn[wnn] = _D[:, 1:], _mnn[:, 1:]
                    
                    #Get d metrics
                    d[wnn] = np.prod(D[wnn], axis=1)
                    d[ideal_idx] = np.inf
                
                #New item removed
                _iter = _iter + 1
                
                #Break by excessive calls
                if _iter > 1e4:
                    print("Breaking max iter in MNN")
                    break
            
            #Initialize as zeros
            d2 = np.zeros(n_points)
            d2[is_unique] = d

            return d2
            

def calc_crowding_entropy(F, filter_out_duplicates=True, **kwargs):
    """Wang, Y.-N., Wu, L.-H. & Yuan , X.-F., 2010. Multi-objective self-adaptive differential evolution with elitist archive and crowding entropy-based diversity measure. Soft Comput., 14(3), pp. 193-209.

    Args:
        F (2d array like): Objective functions.
        filter_out_duplicates (bool, optional): Defaults to True.

    Returns:
        crowding_enropies (1d array)
    """
    n_points, n_obj = F.shape

    if n_points <= 2:
        return np.full(n_points, np.inf)

    else:

        if filter_out_duplicates:
            # filter out solutions which are duplicates - duplicates get a zero finally
            is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-32)))[0]
        else:
            # set every point to be unique without checking it
            is_unique = np.arange(n_points)

        # index the unique points of the array
        _F = F[is_unique]

        # sort each column and get index
        I = np.argsort(_F, axis=0, kind='mergesort')

        # sort the objective space values for the whole matrix
        _F = _F[I, np.arange(n_obj)]

        # calculate the distance from each point to the last and next
        dist = np.row_stack([_F, np.full(n_obj, np.inf)]) - np.row_stack([np.full(n_obj, -np.inf), _F])

        # calculate the norm for each objective - set to NaN if all values are equal
        norm = np.max(_F, axis=0) - np.min(_F, axis=0)
        norm[norm == 0] = np.nan

        # prepare the distance to last and next vectors
        dl = dist.copy()[:-1]
        du = dist.copy()[1:]
        
        #Fix nan
        dl[np.isnan(dl)] = 0.0
        du[np.isnan(du)] = 0.0
        
        #Total distance
        cd = dl + du

        #Get relative positions
        pl = (dl[1:-1] / cd[1:-1])
        pu = (du[1:-1] / cd[1:-1])

        #Entropy
        entropy = np.row_stack([np.full(n_obj, np.inf),
                                -(pl * np.log2(pl) + pu * np.log2(pu)),
                                np.full(n_obj, np.inf)])
        
        #Crowding entropy
        J = np.argsort(I, axis=0)
        _cej = cd[J, np.arange(n_obj)] * entropy[J, np.arange(n_obj)] / norm
        _cej[np.isnan(_cej)] = 0.0
        _ce = _cej.sum(axis=1)

        #Save the final vector which sets the crowding distance for duplicates to zero to be eliminated
        ce = np.zeros(n_points)
        ce[is_unique] = _ce

    return ce


def _calc_mnn(F, filter_out_duplicates=True, **kwargs):
    n_points, n_obj = F.shape

    if n_points <= 2:
        return np.full(n_points, np.inf)

    else:

        if filter_out_duplicates:
            # filter out solutions which are duplicates - duplicates get a zero finally
            is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-32)))[0]
        else:
            # set every point to be unique without checking it
            is_unique = np.arange(n_points)

        # index the unique points of the array
        _F = F[is_unique].copy()

        # calculate the norm for each objective - set to NaN if all values are equal
        norm = np.max(_F, axis=0) - np.min(_F, axis=0)
        norm[norm == 0] = 1.0
        
        # F normalized
        _F = (_F - _F.min(axis=0)) / norm
        
        # Distances pairwise (Inefficient)
        D = pdist(_F, metric="euclidean")
        D = squareform(D)
        
        # M neighbors
        M = _F.shape[1]
        _D = np.sort(D, axis=1)[:, 1:M+1]
        
        # Metric d
        _d = np.prod(_D, axis=1)
        
        # Set top performers as np.inf
        _extremes = np.argmin(_F, axis=0)
        _d[_extremes] = np.inf

        # Save the final vector which sets the crowding distance for duplicates to zero to be eliminated
        d = np.zeros(n_points)
        d[is_unique] = _d

    return d



import numpy as np
from pymoo.core.crossover import Crossover
from pymoo.core.population import Population
from pymoo.operators.crossover.binx import mut_binomial
from pymoo.operators.crossover.expx import mut_exp


# =========================================================================================================
# Implementation
# =========================================================================================================

class DEM:
    
    def __init__(self,
                 F=None,
                 gamma=1e-4,
                 repair="bounce-back",
                 **kwargs):

        #Default value for F
        if F is None:
            F = (0.0, 1.0)
        
        #Define which method will be used to generate F values
        if hasattr(F, "__iter__"):
            self.scale_factor = self._randomize_scale_factor
        else:
            self.scale_factor = self._scalar_scale_factor
        
        #Define which method will be used to generate F values
        if not hasattr(repair, "__call__"):
            try:
                repair = REPAIRS[repair]
            except:
                raise KeyError("Repair must be either callable or in " + str(list(REPAIRS.keys())))
        
        #Define which strategy of rotation will be used
        if gamma is None:
            self.get_diff = self._diff_simple
        else:
            self.get_diff = self._diff_jitter
        
        self.F = F
        self.gamma = gamma
        self.repair = repair
        
    def do(self, problem, pop, parents, **kwargs):

        #Get all X values for mutation parents
        Xr = pop.get("X")[parents.T].copy()
        assert len(Xr.shape) == 3, "Please provide a three-dimensional matrix n_parents x pop_size x n_vars."
        
        #Create mutation vectors
        V, diffs = self.de_mutation(Xr, return_differentials=True)

        #If the problem has boundaries to be considered
        if problem.has_bounds():
            
            #Do repair
            V = self.repair(V, Xr[0], *problem.bounds())
                
        return Population.new("X", V)
    
    def de_mutation(self, Xr, return_differentials=True):
        
        n_parents, n_matings, n_var = Xr.shape
        assert n_parents % 2 == 1, "For the differential an odd number of values need to be provided"

        #Build the pairs for the differentials
        pairs = (np.arange(n_parents - 1) + 1).reshape(-1, 2)

        #The differentials from each pair subtraction
        diffs = self.get_diffs(Xr, pairs, n_matings, n_var)

        #Add the difference vectors to the base vector
        V = Xr[0] + diffs

        if return_differentials:
            return V, diffs
        else:
            return V
    
    def _randomize_scale_factor(self, n_matings):
         return (self.F[0] + np.random.random(n_matings) * (self.F[1] - self.F[0]))
     
    def _scalar_scale_factor(self, n_matings):
         return np.full(n_matings, self.F)
     
    def _diff_jitter(self, F, Xi, Xj, n_matings, n_var):
        F = F[:, None] * (1 + self.gamma * (np.random.random((n_matings, n_var)) - 0.5))
        return F * (Xi - Xj)
    
    def _diff_simple(self, F, Xi, Xj, n_matings, n_var):
        return F[:, None] * (Xi - Xj)
    
    def get_diffs(self, Xr, pairs, n_matings, n_var):
        
        #The differentials from each pair subtraction
        diffs = np.zeros((n_matings, n_var))
        
        #For each difference
        for i, j in pairs:
        
            #Obtain F randomized in range
            F = self.scale_factor(n_matings)
            
            #New difference vector
            diff = self.get_diff(F, Xr[i], Xr[j], n_matings, n_var)
            
            #Add the difference to the first vector
            diffs = diffs + diff
        
        return diffs
        
    
class DEX(Crossover):
    
    def __init__(self,
                 variant="bin",
                 CR=0.7,
                 F=None,
                 gamma=1e-4,
                 n_diffs=1,
                 at_least_once=True,
                 repair="bounce-back",
                 **kwargs):
        
        #Default value for F
        if F is None:
            F = (0.0, 1.0)
        
        #Create instace for mutation
        self.dem = DEM(F=F,
                       gamma=gamma,
                       repair=repair)
    
        self.CR = CR
        self.variant = variant
        self.at_least_once = at_least_once
        
        super().__init__(2 + 2 * n_diffs, 1,  prob=1.0, **kwargs)
    
    def do(self, problem, pop, parents, **kwargs):
        
        #Get target vectors
        X = pop.get("X")[parents[:, 0]]
        
        #About Xi
        n_matings, n_var = X.shape
        
        #Obtain mutants
        mutants = self.dem.do(problem, pop, parents[:, 1:], **kwargs)
        
        #Obtain V
        V = mutants.get("X")
        
        #Binomial crossover
        if self.variant == "bin":
            M = mut_binomial(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        #Exponential crossover
        elif self.variant == "exp":
            M = mut_exp(n_matings, n_var, self.CR, at_least_once=self.at_least_once)
        else:
            raise Exception(f"Unknown variant: {self.variant}")

        #Add mutated elements in corresponding main parent
        X[M] = V[M]

        off = Population.new("X", X)
        
        return off
    

def bounce_back(X, Xb, xl, xu):
    """Repair strategy

    Args:
        X (2d array like): Mutated vectors including violations.
        Xb (2d array like): Reference vectors for repair in feasible space.
        xl (1d array like): Lower-bounds.
        xu (1d array like): Upper-bounds.

    Returns:
        2d array like: Repaired vectors.
    """
    
    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - Xb[i, j])

    return X

def midway(X, Xb, xl, xu):
    """Repair strategy

    Args:
        X (2d array like): Mutated vectors including violations.
        Xb (2d array like): Reference vectors for repair in feasible space.
        xl (1d array like): Lower-bounds.
        xu (1d array like): Upper-bounds.

    Returns:
        2d array like: Repaired vectors.
    """
    
    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + (Xb[i, j] - XL[i, j]) / 2

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - (XU[i, j] - Xb[i, j]) / 2

    return X

def to_bounds(X, Xb, xl, xu):
    """Repair strategy

    Args:
        X (2d array like): Mutated vectors including violations.
        Xb (2d array like): Reference vectors for repair in feasible space.
        xl (1d array like): Lower-bounds.
        xu (1d array like): Upper-bounds.

    Returns:
        2d array like: Repaired vectors.
    """
    
    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j]

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j]

    return X

def rand_init(X, Xb, xl, xu):
    """Repair strategy

    Args:
        X (2d array like): Mutated vectors including violations.
        Xb (2d array like): Reference vectors for repair in feasible space.
        xl (1d array like): Lower-bounds.
        xu (1d array like): Upper-bounds.

    Returns:
        2d array like: Repaired vectors.
    """
    
    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    return X


def squared_bounce_back(X, Xb, xl, xu):
    """Repair strategy

    Args:
        X (2d array like): Mutated vectors including violations.
        Xb (2d array like): Reference vectors for repair in feasible space.
        xl (1d array like): Lower-bounds.
        xu (1d array like): Upper-bounds.

    Returns:
        2d array like: Repaired vectors.
    """
    
    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.square(np.random.random(len(i))) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.square(np.random.random(len(i))) * (XU[i, j] - Xb[i, j])

    return X

def normalize_fun(fun):
    
    fmin = fun.min(axis=0)
    fmax = fun.max(axis=0)
    den = fmax - fmin
    
    den[den <= 1e-16] = 1.0
    
    return (fun - fmin)/den

REPAIRS = {"bounce-back":bounce_back,
           "midway":midway,
           "rand-init":rand_init,
           "to-bounds":to_bounds}



"""Created by

Bruno Scalia C. F. Leite, 2021

"""

import numpy as np
from pymoo.core.selection import Selection


# =========================================================================================================
# Implementation
# =========================================================================================================


#This is the core differential evolution selection class
class DES(Selection):

    def __init__(self,
                 variant,
                 **kwargs):
        
        super().__init__()
        self.variant = variant

    def _do(self, pop, n_select, n_parents, **kwargs):
        
        #Obtain number of elements in population
        n_pop = len(pop)
        
        #For most variants n_select must be equal to len(pop)
        variant = self.variant
            
        if variant == "ranked":
            """Proposed by Zhang et al. (2021). doi.org/10.1016/j.asoc.2021.107317"""
            P = self._ranked(pop, n_select, n_parents)
        
        elif variant == "best":
            P = self._best(pop, n_select, n_parents)
        
        elif variant == "current-to-best":
            P = self._current_to_best(pop, n_select, n_parents)
        
        elif variant == "current-to-rand":
            P = self._current_to_rand(pop, n_select, n_parents)
            
        else:
            P = self._rand(pop, n_select, n_parents)

        return P
    
    def _rand(self, pop, n_select, n_parents, **kwargs):
        
        #len of pop
        n_pop = len(pop)

        #Base form
        P = np.empty([n_select, n_parents], dtype=int)
        
        #Fill first column with corresponding parent
        P[:, 0] = np.arange(n_pop)

        #Fill next columns in loop
        for j in range(1, n_parents):
            
            P[:, j] = np.random.choice(n_pop, n_select)            
            reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
            
            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
        
        return P
    
    def _best(self, pop, n_select, n_parents, **kwargs):
        
        #len of pop
        n_pop = len(pop)

        #Base form
        P = np.empty([n_select, n_parents], dtype=int)
        
        #Fill first column with corresponding parent
        P[:, 0] = np.arange(n_pop)
        
        #Fill first column with best candidate
        P[:, 1] = 0

        #Fill next columns in loop
        for j in range(2, n_parents):
            
            P[:, j] = np.random.choice(n_pop, n_select)            
            reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
            
            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
        
        return P
    
    def _current_to_best(self, pop, n_select, n_parents, **kwargs):
        
        #len of pop
        n_pop = len(pop)

        #Base form
        P = np.empty([n_select, n_parents], dtype=int)
        
        #Fill first column with corresponding parent
        P[:, 0] = np.arange(n_pop)
        
        #Fill first column with current candidate
        P[:, 1] = np.arange(n_pop)
        
        #Fill first direction from current
        P[:, 3] = np.arange(n_pop)
        
        #Towards best
        P[:, 2] = 0

        #Fill next columns in loop
        for j in range(4, n_parents):
            
            P[:, j] = np.random.choice(n_pop, n_select)            
            reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
            
            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
        
        return P
    
    def _current_to_rand(self, pop, n_select, n_parents, **kwargs):
        
        #len of pop
        n_pop = len(pop)

        #Base form
        P = np.empty([n_select, n_parents], dtype=int)
        
        #Fill first column with corresponding parent
        P[:, 0] = np.arange(n_pop)
        
        #Fill first column with current candidate
        P[:, 1] = np.arange(n_pop)
        
        #Fill first direction from current
        P[:, 3] = np.arange(n_pop)
        
        #Towards random
        P[:, 2] = np.random.choice(n_pop, n_select)            
        reselect = (P[:, 2].reshape([-1, 1]) == P[:, [0, 1, 3]]).any(axis=1)
            
        while np.any(reselect):
            P[reselect, 2] = np.random.choice(n_pop, reselect.sum())
            reselect = (P[:, 2].reshape([-1, 1]) == P[:, [0, 1, 3]]).any(axis=1)

        #Fill next columns in loop
        for j in range(4, n_parents):
            
            P[:, j] = np.random.choice(n_pop, n_select)            
            reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
            
            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)
        
        return P
    
    def _ranked(self, pop, n_select, n_parents, **kwargs):
        
        P = self._rand(pop, n_select, n_parents, **kwargs)
        P[:, 1:] = rank_sort(P[:, 1:], pop)
        
        return P
    

def ranks_from_cv(pop):
    
    ranks = pop.get("rank")
    cv_elements = ranks == None
    
    if np.any(cv_elements):
        ranks[cv_elements] = np.arange(len(pop))[cv_elements]
    
    return ranks

def rank_sort(P, pop):
    
    ranks = ranks_from_cv(pop)
    
    sorted = np.argsort(ranks[P], axis=1, kind="stable")    
    S = np.take_along_axis(P, sorted, axis=1)

    P[:, 0] = S[:, 0]
    
    n_diffs = int((P.shape[1] - 1) / 2)

    for j in range(1, n_diffs + 1):
        P[:, 2*j - 1] = S[:, j]
        P[:, 2*j] = S[:, -j]
    
    return P

def reiforce_directions(P, pop):
    
    ranks = ranks_from_cv(pop)
    
    ranks = ranks[P] 
    S = P.copy()
    
    n_diffs = int(P.shape[1] / 2)

    for j in range(0, n_diffs):
        bad_directions = ranks[:, 2*j] > ranks[:, 2*j + 1]
        P[bad_directions, 2*j] = S[bad_directions, 2*j + 1]
        P[bad_directions, 2*j + 1] = S[bad_directions, 2*j]
    
    return P



import numpy as np
from sklearn.neighbors import NearestNeighbors
from pymoo.core.indicator import Indicator

class SpacingIndicator(Indicator):
    
    def __init__(self,
                 zero_to_one=False,
                 ideal=None,
                 nadir=None):
        
        super().__init__(zero_to_one=zero_to_one,
                         ideal=ideal,
                         nadir=nadir)
    
    def _do(self, F, *args, **kwargs):
        
        #Get F dimensions
        n_points, n_obj = F.shape
        
        #knn
        knn = NearestNeighbors(n_neighbors=2, metric="manhattan")
        knn.fit(F)
        
        #Get distances (nearest neighbor is the point itself)
        _d, _ = knn.kneighbors(F.copy(), return_distance=True)
        d = _d[:, -1]
        dm = np.mean(d)
        
        #Get spacing
        S = np.sqrt(np.sum(np.square(d - dm)) / n_points)
        
        return S


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
  name = 'pymoode',
  packages = ['pymoode'],
  version = '0.1.7',
  license='Apache License 2.0',
  description = 'A Python optimization package using Differential Evolution.',
  long_description=long_description,
  long_description_content_type="text/markdown",
  author = 'Bruno Scalia C. F. Leite',
  author_email = 'mooscaliaproject@gmail.com',
  url = 'https://github.com/mooscaliaproject/pymoode',
  download_url = 'https://github.com/mooscaliaproject/pymoode',
  keywords = ['Multi-objective optimization',
              'NSGA-II',
              'GDE3',
              'NSDE',
              'SA-NSDE',
              'NSDE-R',
              'Differential Evolution',
              'Genetic Algorithm',
              'Evolutionary Algorithms',
              'Evolutionary optimization'],
  install_requires=[
          'numpy>=1.19.*',
          'pymoo==0.5.*',
          'scikit-learn>=1.0.*',
          'scipy>=1.7.*',
      ],
)


"""Created by

Bruno Scalia C. F. Leite, 2022

"""

from pymoode.nsde import NSDE
from pymoo.core.population import Population
from pymoo.util.dominator import get_relation


# =========================================================================================================
# Implementation
# =========================================================================================================


class GDE3(NSDE):
    
    def __init__(self,
                 pop_size=100,
                 variant="DE/rand/1/bin",
                 CR=0.5,
                 F=None,
                 gamma=1e-4,
                 **kwargs):
        """
        GDE3 is an extension of DE to multi-objective problems using a mixed type survival strategy.
        It is implemented in this version with the same constraint handling strategy of NSGA-II by default.
        
        For many-objective problems, try using NSDE-R or RankSurvival with 'mnn' crowding metric.

        Kukkonen, S. & Lampinen, J., 2005. GDE3: The third evolution step of generalized differential evolution. 2005 IEEE congress on evolutionary computation, Volume 1, pp. 443-450.

        Parameters
        ----------
        
        pop_size : int, optional
            Population size. Defaults to 100.
            
        sampling : Sampling, optional
            Sampling strategy of pymoo. Defaults to LHS().
            
        variant : str, optional
            Differential evolution strategy. Must be a string in the format:
            "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'.
            Selection variants are:
            
                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'
                
            The selection strategy 'ranked' might be helpful to improve convergence speed without much harm to diversity. Defaults to 'DE/rand/1/bin'.
            
        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.
            
        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.
            
        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.
            
        pm : Mutation, optional
            Pymoo's mutation operators after crossover. Defaults to NoMutation().
            
        reapair : Repair, optional
            Repair of mutant vectors. Is either callable or one of:
        
                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'
            
            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations, Xb contains reference vectors for repair in feasible space, xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.
            
        survival : Survival, optional
            Pymoo's survival strategy.
            Defaults to RankSurvival() with bulk removal ('full') and crowding distances ('cd').
            In GDE3, the survival strategy is applied after a one-to-one comparison between child vector and corresponding parent when both are non-dominated by the other.
        """
        
        super().__init__(pop_size=pop_size,
                         variant=variant,
                         CR=CR,
                         F=F,
                         gamma=gamma,
                         **kwargs)

    def _advance(self, infills=None, **kwargs):
        
        assert infills is not None, "This algorithms uses the AskAndTell interface thus 'infills' must to be provided."

        #The individuals that are considered for the survival later and final survive
        survivors = []

        # now for each of the infill solutions
        for k in range(len(self.pop)):

            #Get the offspring an the parent it is coming from
            off, parent = infills[k], self.pop[k]

            #Check whether the new solution dominates the parent or not
            rel = get_relation(parent, off)

            #If indifferent we add both
            if rel == 0:
                survivors.extend([parent, off])

            #If offspring dominates parent
            elif rel == -1:
                survivors.append(off)

            #If parent dominates offspring
            else:
                survivors.append(parent)

        #Create the population
        survivors = Population.create(*survivors)

        #Perform a survival to reduce to pop size
        self.pop = self.survival.do(self.problem, survivors, n_survive=self.n_offsprings)
